/*
 *      File Name     : StdoutMonitor.java
 *
 *      Product       : CICS Transaction Gateway
 *
 *      Description   : This sample shows the basic use of the CICS
 *                      Transaction Gateway request monitoring exits. 
 *                      This will simply output the data available at 
 *                      each exit point to the given "out" file 
 *                      (or stdout if not provided).
 *                      "out" is defined by the system property:
 *                      com.ibm.ctg.samples.requestexit.out
 *                      
 *      Pre-Requisites: Use a version of the JDK that the CICS Transaction
 *                      Gateway supports if you recompile this sample. See
 *                      the CICS Transaction Gateway Administration book for
 *                      supported Java levels.
 *
 * @start_copyright_sample_CTG_ZOS@
 * Licensed Materials - Property of IBM
 * 
 * 5724-I81 5655-R25
 * 
 * (c) Copyright IBM Corp. 2007, 2008
 * 
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with
 * IBM Corp.
 * 
 * Status: Version 7.1.0.2
 * @end_copyright_sample_CTG_ZOS@
 *
 *      The following code is sample code created by IBM Corporation.  This
 *      sample code is not part of any standard IBM product and is provided
 *      to you solely for the purpose of assisting you in the development of
 *      your applications.  The code is provided 'AS IS', without warranty
 *      or condition of any kind.  IBM shall not be liable for any damages
 *      arising out of your use of the sample code, even if IBM has been
 *      advised of the possibility of such damages.
 */

package com.ibm.ctg.samples.requestexit;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.EnumSet;
import java.util.Map;

import javax.transaction.xa.Xid;

import com.ibm.ctg.monitoring.FlowTopology;
import com.ibm.ctg.monitoring.FlowType;
import com.ibm.ctg.monitoring.RequestData;
import com.ibm.ctg.monitoring.RequestEvent;
import com.ibm.ctg.monitoring.RequestExit;
import com.ibm.ctg.monitoring.TransientPayLoad;
import com.ibm.ctg.util.OriginData;

/**
 * The StdoutMonitor is designed to output all the data available
 * on the eventFired method.  It is written so subclasses can re-use
 * the output methods.  Output will be written to System.out, unless
 * overridden by the system property defined by
 * {@link com.ibm.ctg.samples.requestexit.StdoutMonitor#STDOUT_PROPERTY STDOUT_PROPERTY}
 */
public class StdoutMonitor implements RequestExit {

   // @start_copyright_object_java_sample_CTG_ZOS@
   public static final String copyright = "Licensed Materials - Property of IBM "+
   "5724-I81 5655-R25 " +
   "(c) Copyright IBM Corp. 2001, 2008  All Rights Reserved. " +
   "US Government Users Restricted Rights - Use, duplication or " +
   "disclosure restricted by GSA ADP Schedule Contract with " +
   "IBM Corp.";
   // @end_copyright_object_java_sample_CTG_ZOS@

   /**
    * The Java property name used to set the output file name.
    */
   public static final String STDOUT_PROPERTY = "com.ibm.ctg.samples.requestexit.out";

   private PrintStream stdout;
   private String prefix;
   
   public boolean onZos;

   /**
    * Default constructor initializes the output methods.
    */
   public StdoutMonitor() {
      // It is useful to remember if we are on z/OS...
      onZos = System.getProperty("os.name").equals("z/OS");
      
      // set up output destination
      String destination = System.getProperty(STDOUT_PROPERTY);
      if (destination == null){
         stdout = getDefaultStdout();
      } else {
         File destFile = new File(destination);
         try {
            stdout = new PrintStream(destFile);
         } catch (FileNotFoundException e) {
            stdout = getDefaultStdout();
            System.err.println("Print to '" + destination + "' failed with " + e);
            System.err.println("Using default PrintStream");
         }
      }
      println("Starting " + getClass().getSimpleName() + 
              " log stream  at " + new Date());
      
      // now output some environment status, just for the record...
      println(getSystemEnvironment());
   }

   /* (non-Javadoc)
    * @see com.ibm.ctg.monitoring.RequestExit#eventFired(com.ibm.ctg.monitoring.RequestEvent, java.util.Map)
    */
   public void eventFired(RequestEvent event, Map<RequestData, Object> data) {
      // log detailed event data for a request/response flow
      if (event.equals(RequestEvent.RequestEntry)
            || event.equals(RequestEvent.ResponseExit)) {
         // use dumpEventData() to format a detailed summary and log it 
         println(dumpEventData(myName() + ":eventFired", event, data));
      } else {
         // log non-request/response events with current time stamp
         Date now = new Date();
         // build common log entry prefix
         StringBuffer message = new StringBuffer(now.toString());
         message.append(" ");
         message.append(myName());
         message.append(":eventFired called with event = ");
         message.append(event.toString());

         // append event detail
         if (event.equals(RequestEvent.Command)) {
            // log the command event and data
            message.append(", Data = ");
            message.append((String) data.get(RequestData.CommandData));
         } else if (event.equals(RequestEvent.ShutDown)) {
            // log the ShutDown event only (no resources need to be freed)
            message.append(" OK");
         } else {
            // log any unknown event types
            message.append(", Unknown event type.");
         }
         println(message.toString());
      }
   }

   /**
    * This method can be overridden by subclasses to give a name to identify the
    * output.
    * 
    * @return defaults to the name of the current class
    */
   public String myName() {
      return this.getClass().getCanonicalName();
   }

   /**
    * This method can be overridden by subclasses to use
    * a different default PrintStream.
    * 
    * @return PrintStream to use as default output device
    */
   public PrintStream getDefaultStdout() {
      return System.out;
   }
   
   /**
    * Method to format all the information available at 
    * an event.
    * 
    * @param originator Text to identify originator
    * @param event that fired
    * @param data found on that event
    * @return User readable dump of exit info
    */
   public String dumpEventData(String originator, RequestEvent event, 
         Map<RequestData, Object> data) {
      /*
       * Build message as one string buffer so seperate threads
       * will not interleave the lines of data.
       */
      int ctgCorelator = (Integer) data.get(RequestData.CtgCorrelator);
      setPrefix(ctgCorelator);
      
      // collect all values not used
      EnumSet<RequestData> remainingData = EnumSet.allOf(RequestData.class);

       // start the header with event details
      StringBuffer message = new StringBuffer();
      appendPrefix(message);
      message.append(originator);
      message.append(" called with event = ");
      message.append(event);
      
      // next add flow description
      FlowType flowType = (FlowType) data.get(RequestData.FlowType);
      appendRequestData(message, RequestData.FlowType, flowType);
      remainingData.remove(RequestData.FlowType);
      FlowTopology flowTopology = (FlowTopology) data.get(RequestData.FlowTopology);
      appendRequestData(message, RequestData.FlowTopology, flowTopology);
      remainingData.remove(RequestData.FlowTopology);
      
      // add Request ID
      Object qualifier = data.get(RequestData.CtgApplidQualifier);
      Object applid = data.get(RequestData.CtgApplid);
      appendFullyQualifiedApplid(message, "Fully qualified applid", qualifier, applid);
      remainingData.remove(RequestData.CtgApplidQualifier);
      remainingData.remove(RequestData.CtgApplid);
      appendRequestData(message, RequestData.CtgCorrelator, ctgCorelator);
      remainingData.remove(RequestData.CtgCorrelator);
      if ( (onZos) &&
           (flowType.equals(FlowType.EciSynconreturn)) &&
           (flowTopology.equals(FlowTopology.Gateway)) &&
           (data.get(RequestData.OriginData) == null) &&
           (data.get(RequestData.Xid) == null) &&
           (data.get(RequestData.LuwToken) == null) ){
         /*
          * If we are on z/OS with a sync on return flow on
          * a Gateway daemon and we do not have Origin Data we 
          * can assume that we are using EXCI.  If we do not
          * have an XA transaction and we are not part of an
          * LUW we will be using an EXCI Network UOW.
          */
         appendNetworkUow(message, qualifier, applid, ctgCorelator);
      }

      // add Client Request ID if present
      Object dataItem = data.get(RequestData.ClientCtgCorrelator); 
      if (dataItem != null){
         qualifier = data.get(RequestData.ClientCtgApplidQualifier);
         applid = data.get(RequestData.ClientCtgApplid);
         appendFullyQualifiedApplid(message, "Client fully qualified applid", qualifier, applid);
         appendRequestData(message, RequestData.ClientCtgCorrelator, dataItem);
      }
      remainingData.remove(RequestData.ClientCtgApplidQualifier);
      remainingData.remove(RequestData.ClientCtgApplid);
      remainingData.remove(RequestData.ClientCtgCorrelator);

      // add XID if present
      dataItem = data.get(RequestData.Xid); 
      if (dataItem != null){
         appendXid(message, dataItem);
      }
      remainingData.remove(RequestData.Xid);

      // add URID if present
      dataItem = data.get(RequestData.Urid); 
      if (dataItem != null){
         appendRequestData(message, RequestData.Urid, dataItem);
      }
      remainingData.remove(RequestData.Urid);

      // add LUW if present
      dataItem = data.get(RequestData.LuwToken); 
      if (dataItem != null){
         appendRequestData(message, RequestData.LuwToken, dataItem);
      }
      remainingData.remove(RequestData.LuwToken);

      // add Origin Data if present
      dataItem = data.get(RequestData.OriginData); 
      if (dataItem != null){
         appendOriginData(message, dataItem);
      }
      remainingData.remove(RequestData.OriginData);

      // add all timestamps that are present
      // get base for the timestamp offsets 
      long base = (Long) data.get(RequestData.RequestReceived);
      appendTimeStamp(message, base, RequestData.RequestReceived, base, 1);
      remainingData.remove(RequestData.RequestReceived);
      dataItem = data.get(RequestData.RequestSent); 
      if (dataItem != null){
         appendTimeStamp(message, base, RequestData.RequestSent, dataItem, 2);
      }
      remainingData.remove(RequestData.RequestSent);
      dataItem = data.get(RequestData.ResponseReceived); 
      if (dataItem != null){
         appendTimeStamp(message, base, RequestData.ResponseReceived, dataItem, 3);
      }
      remainingData.remove(RequestData.ResponseReceived);
      dataItem = data.get(RequestData.ResponseSent); 
      if (dataItem != null){
         appendTimeStamp(message, base, RequestData.ResponseSent, dataItem, 4);
      }
      remainingData.remove(RequestData.ResponseSent);

      // output any remaining data fields in a consistent order
      for (RequestData requestData: remainingData){
         Object object = data.get(requestData);
         // only output records for data types that are present
         if (object != null){
            if (requestData == RequestData.PayLoad) {
               appendPayLoad(message, requestData, object);
            } else {
               appendRequestData(message, requestData, object);
            }
         }
      }
      return message.toString();
   }

   /**
    * The PayLoad is provided so that embedded correlators used by
    * 3rd party applications can be used in the monitoring exits.
    * As an example the first and last 32 bytes are displayed.
    *  
    * @param message
    * @param requestData
    * @param object
    */
   private void appendPayLoad(StringBuffer message, RequestData requestData, Object object) {
      TransientPayLoad tpl = (TransientPayLoad) object;
      appendPrefix(message);
      message.append(requestData);
      message.append(" = Commarea is ");
      message.append(tpl.getLength());
      message.append(" long");
      byte[] first = tpl.getBytes(0, 32);
      appendPrefix(message);
      message.append("   First 32 bytes: ");
      message.append(dumpToHex(first));
      byte[] last = tpl.getBytes(0, -32);
      appendPrefix(message);
      message.append("   Last 32 bytes:  ");
      message.append(dumpToHex(last));
   }

   /**
    * As most of the Origin Data is already included in other
    * RequestData fields, just outoput the Transaction Group ID.
    * 
    * @param message
    * @param dataItem
    */
   private void appendOriginData(StringBuffer message, Object dataItem) {
      // most of the origin data is included in other fields
      appendPrefix(message);
      message.append(RequestData.OriginData);
      message.append(" - Transaction Group ID = ");
      OriginData originData = (OriginData) dataItem;
      message.append(toHex(originData.getTransGroupId()));
   }

   /**
    * Format the XID into the constituent parts: Format ID, 
    * Branch Qualifier and GTRID.
    * 
    * @param message
    * @param dataItem
    */
   private void appendXid(StringBuffer message, Object dataItem) {
      Xid xid = (Xid) dataItem;
      appendPrefix(message);
      message.append(RequestData.Xid);
      message.append(" = Format ID: ");
      message.append(Integer.toHexString(xid.getFormatId()));
      appendPrefix(message);
      message.append("   BQUAL: ");
      message.append(toHex(xid.getBranchQualifier()));
      appendPrefix(message);
      message.append("   GTRID: ");
      message.append(toHex(xid.getGlobalTransactionId()));
   }

   /**
    * Append the timestamp in a user readable form.
    * We add the T1 - T4 tags used in the javadoc to describe
    * the location of the timestamps in the flow.
    * Timestamps are formatted as date / time and as an offset
    * in milliseconds from the time the request is first receiced.
    *  
    * @param message
    * @param base
    * @param requestData
    * @param dataItem
    * @param tFlag
    */
   private void appendTimeStamp(StringBuffer message, long base, RequestData requestData, Object dataItem, int tFlag) {
      Long value = (Long)dataItem;
      Date timestamp = new Date(value);
      appendPrefix(message);
      message.append("T");
      message.append(tFlag);
      message.append(" - ");
      message.append(requestData);
      message.append(" (");
      message.append(value);
      message.append(") \t= ");
      message.append(timestamp);
      if (tFlag > 1){
         message.append(" offset = ");
         message.append((value.longValue() - base));
      }
   }

   /**
    * The fully qualified applid is made up of the 2 CICS TG fields used 
    * identify the Gateway daemon or Gateway classes.
    * 
    * @param qualifier
    * @param applid
    */
   public void appendFullyQualifiedApplid(StringBuffer message, String name, Object qualifier, Object applid) {
      String requestId = "";
      if (qualifier != null){
         requestId = (String) qualifier;
      }
      if (applid != null){
         if (qualifier != null){
            requestId += ".";
         }
         requestId += (String) applid;
      } else {
         if (qualifier == null){
            requestId = "NoApplid";
         }
      }
      appendPrefix(message);
      message.append(name);
      message.append(" = ");
      message.append(requestId);
   }

   /**
    * The pseudo EXCI Network Unit of Work is made up of the 3 CICS TG 
    * fields used to correlate a request between RequestEvents.
    * If CtgApplid is not defined then the JNI will use the actual
    * job name for the applid.  This is not available in the exit, so
    * we will use "JOBNAME" as the job name in this case. If CtgApplid 
    * is provided then the pseudo EXCI Network Unit of Work is the real
    * EXCI Network Unit of Work used.
    * 
    * @param qualifier
    * @param applid
    * @param correlator
    * @param addExciUowid
    */
   public void appendNetworkUow(StringBuffer message, Object qualifier, Object applid, int correlator) {
      boolean pseudo = false;
      String requestId = "";
      if (qualifier != null){
         requestId = (String) qualifier;
      }
      if (applid == null){
         applid = "JOBNAME";
         pseudo = true;
      }
      if (qualifier != null){
         requestId += ".";
      }
      requestId += (String) applid;
      byte[] ebcdicUowid = null;
      try {
         ebcdicUowid = requestId.getBytes("IBM-037");
      } catch (UnsupportedEncodingException e) {
         // if iconv fails use EBCDIC version of "Error"
         ebcdicUowid = new byte[] {
               (byte)0xc5, (byte)0xd9, (byte)0xd9, (byte)0xd6, (byte)0xd9
         };
      }
      byte nameLength = (byte) ebcdicUowid.length;
      byte wholeLength = (byte) (nameLength + 1 + 8);
      byte[] formattedUowid = new byte[wholeLength + 1];
      formattedUowid[0] = wholeLength; 
      formattedUowid[1] = nameLength;
      System.arraycopy(ebcdicUowid, 0, formattedUowid, 2, nameLength);
      formattedUowid[2 + nameLength + 0] = (byte) (correlator >> 24);
      formattedUowid[2 + nameLength + 1] = (byte) (correlator >> 16);
      formattedUowid[2 + nameLength + 2] = (byte) (correlator >> 8);
      formattedUowid[2 + nameLength + 3] = (byte) (correlator);
      formattedUowid[2 + nameLength + 4] = 0;
      formattedUowid[2 + nameLength + 5] = 0;
      formattedUowid[2 + nameLength + 6] = 0;
      formattedUowid[2 + nameLength + 7] = 0;
      appendPrefix(message);
      if (pseudo){
         message.append("Pseudo ");
      }
      message.append("EXCI Network UOWID = ");
      message.append(toHex(formattedUowid));
   }

   /**
    * Append the RequestData using simple formatting.
    * 
    * @param message
    * @param requestData
    * @param data
    */
   public void appendRequestData(StringBuffer message, RequestData requestData, Map<RequestData, Object> data) {
      appendRequestData(message, requestData, data.get(requestData));
   }

   /**
    * Append the RequestData using simple formatting.
    * 
    * @param message
    * @param requestData
    * @param object
    */
   public void appendRequestData(StringBuffer message, RequestData requestData, Object object) {
      appendPrefix(message);
      message.append(requestData);
      message.append(" = ");
      if (object instanceof byte[]) {
         byte[] bytes = (byte[]) object;
         message.append(toHex(bytes));
      } else {
         message.append(object);
      }
   }

   /**
    * Append the prefix string to the message buffer.
    * 
    * @param message StringBuffer to append to.
    */
   public void appendPrefix(StringBuffer message) {
      message.append(prefix);
   }

   /**
    * Set the prefix for each output line using the
    * current CTG Correlator.
    * 
    * @param ctgCorelator to use for prefix
    */
   public void setPrefix(int ctgCorelator) {
      prefix = String.format("\n[%011d]: ", ctgCorelator);
   }

   /**
    * Simple output routine to send a string to our currently set
    * PrintWriter, appending a newline, and flushing the buffer.
    * 
    * Output will be written to System.out, unless
    * overridden by the system property defined by
    * {@link com.ibm.ctg.samples.requestexit.StdoutMonitor#STDOUT_PROPERTY STDOUT_PROPERTY}.
    * 
    * To ensure the whole message goes in one block it is synchronized. 
    * 
    * @param string to output
    */
   public synchronized void println(String string){
      print(string + "\n");
      if (stdout == null){
         System.out.flush();
      } else {
         stdout.flush();
      }
   }

   /**
    * Simple output routine to send a string to our currently set
    * output PrintWriter.
    * 
    * @param string to output
    */
   public void print(String args) {
      if (stdout == null){
         System.out.print(args);
      } else {
         try {
            stdout.print(args);
         } catch (Throwable t) {
            System.err.println("Error using stdout: " + t);
            if (stdout != getDefaultStdout()){
               stdout = getDefaultStdout();
               System.err.println("Using default PrintStream");
            } else {
               /*
                * as we cannot output anything we cannot do
                * anything more, so Throw something to disable
                * this exit.
                */
               throw new RuntimeException(
                     "Default PrintStream has failed!");
            }
            print(args);
         }
      }
   }

   /**
    * Format an array of bytes into a string containing the
    * bytes as hexadecimal characters and a printable representation of
    * the bytes as a string
    * 
    * @param bytes array to format
    * @return printable string
    */
   public static String dumpToHex(byte[] bytes) {
      String hexString = toHex(bytes);
      if (bytes != null){
         hexString += " '";
         for (int x = 0; x < bytes.length; x++) {
            byte ch = bytes[x];
            if ( (ch < 32) || (ch > 127) ) {
               ch = '?';
            }
            hexString += new String(Character.toChars(ch));
         }
         hexString += "'";
      }
      return hexString;
   }

   /**
    * Method to format an array of bytes into a readable string of
    * hexadecimal characters.
    * 
    * @param bytes
    * @return String representing bytes in hex
    */
   public static String toHex(byte [] bytes){
      String hexString = "";
      if (bytes == null){
         hexString = "{empty}";
      } else {
         for (int x = 0; x < bytes.length; x++) {
            if ( (x > 0) && (x % 4 == 0) ){
               hexString += " ";
            }
            hexString += String.format("%02X", bytes[x] & 0xFF);
         }
      }
      return hexString;
   }
   
   /**
    * Create a multiline string of the significant system properties.
    */
   public String getSystemEnvironment() {
      StringBuffer message = new StringBuffer("System properties :\n");
      appendProperty(message, "java.version");
      appendProperty(message, "java.vendor");
      appendProperty(message, "java.class.path");
      appendProperty(message, "java.class.version");
      appendOsDetails(message); 
      appendProperty(message, "java.library.path");
      appendProperty(message, "user.name");
      appendProperty(message, "user.dir");
      appendProperty(message, "user.language");
      appendProperty(message, "user.timezone");
      appendProperty(message, "file.encoding");
      return message.toString();
   }

   /**
    * Append the property details.
    * Some properties will throw a SecurityException if used on an 
    * untrusted applet.  Do a getproperty before the first append, 
    * so that we get a clean output if it generates a security exception.
    * 
    * @param message
    * @param property
    */
   private void appendProperty(StringBuffer message, String property) {
      try {
         String tmp = System.getProperty(property);
         message.append(property);
         message.append(" = ");
         message.append(tmp);
         message.append('\n');
         tmp=null;
      }catch (SecurityException eSec) {
         message.append("Security Exception getting " + property + "\n");
      }      
   }

   /**
    * Append the details about the perating system.
    * 
    * @param message
    */
   private void appendOsDetails(StringBuffer message) {
      message.append("OS details :- ");
      message.append(System.getProperty("os.name"));
      message.append(' ');
      message.append(System.getProperty("os.version"));
      message.append(' ');
      message.append(System.getProperty("os.arch"));
      message.append('\n');
   }

}
